__all__ = ('SubterraneanHelpCommand', )

from functools import partial as partial_func

from ....backend.utils import DOCS_ENABLED, doc_property
from ....backend.analyzer import CallableAnalyzer
from ....discord.embed import Embed, EmbedBase
from ....discord.utils import chunkify
from ....discord.preconverters import preconvert_str
from ....discord.user import UserBase, User
from ....discord.channel import ChannelBase, ChannelGuildBase, ChannelTextBase, ChannelText, ChannelPrivate, \
    ChannelVoice, ChannelGroup, ChannelCategory, ChannelStore, ChannelThread, ChannelDirectory
from ....discord.role import Role
from ....discord.emoji.emoji import Emoji
from ....discord.guild import Guild
from ....discord.message import Message
from ....discord.invite import Invite

from ...command_utils import Pagination, Closer
from ..content_parser import timedelta, relativedelta
from ..utils import raw_name_to_display
from ..command_helpers import run_checks

MAX_SUB_COMMAND_DEEPNESS = 5

MAX_LINE_PER_PAGE = 24
DEFAULT_PREFIX = '**>>**'

PARAMETER_REQUIRED_START = '<'
PARAMETER_REQUIRED_END = '>'
PARAMETER_OPTIONAL_START = '*'
PARAMETER_OPTIONAL_END = '*'

PARAMETER_NAME_REST = 'rest'

PARAMETER_TYPE_TO_NAME = {
    UserBase: 'user',
    User: 'user',
    ChannelBase: 'channel',
    ChannelGuildBase: 'guild-channel',
    ChannelTextBase: 'any-text-channel',
    ChannelText: 'text-channel',
    ChannelPrivate: 'private-channel',
    ChannelVoice: 'voice-channel',
    ChannelGroup: 'group-channel',
    ChannelCategory: 'category-channel',
    ChannelStore: 'store-channel',
    ChannelDirectory: 'directory-channel',
    ChannelThread: 'thread-channel',
    Invite: 'invite',
    Role: 'role',
    Guild: 'guild',
    Message: 'message',
    str: 'string',
    int: 'integer',
    timedelta: 'time-delta',
    Emoji: 'emoji',
}

if (relativedelta is not None):
    PARAMETER_TYPE_TO_NAME[relativedelta] = 'relative-time-delta'

PARAMETER_UNDEFINED = 'undefined'

DEFAULT_HELP_NAME = 'help'

PARAMETER_SEPARATOR_SPACE_AFTER_ONLY = {
    ',',
    ':',
    ';',
}


def check_user(user, event):
    """
    Checks whether the ``ReactionAddEvent`` or ``ReactionDeleteEvent`` instance's user is same as the given one.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user who should be matched.
    event : ``ReactionAddEvent`` or ``ReactionDeleteEvent``
        The reaction addition or deletion event.
    """
    return (event.user is user)


async def run_command_checks(command_context, command):
    """
    Returns whether the given command checks pass and it is not hidden either.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        Respective command context.
    command : ``Command``
        The command to check.
    
    Returns
    -------
    should_show : `bool`
    """
    if command.hidden:
        return False
    
    if command.hidden_if_checks_fail:
        failed_check = await run_checks(command._iter_own_checks(), command_context)
        if (failed_check is not None):
            return False
    
    return True


async def run_command_checks_in_category(command_context, command):
    """
    Returns whether the given command should be shown. Checks the category and the command as well.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        Respective command context.
    command : ``Command``
        The command to check.
    
    Returns
    -------
    should_show : `bool`
        Whether the command should be shown.
    """
    category = command.get_category()
    if category is None:
        # runtime error
        return False
    
    if category.hidden:
        return False
    
    if category.hidden_if_checks_fail:
        failed_check = await run_checks(category._iter_own_checks(), command_context)
        if (failed_check is not None):
            return False
    
    if command.hidden:
        return False
    
    if command.hidden_if_checks_fail:
        failed_check = await run_checks(command._iter_own_checks(), command_context)
        if (failed_check is not None):
            return False
    
    return True


async def is_category_command_shown(command_context, category):
    """
    Returns whether the any command of the category should be shown.
    
    This function is a coroutine.
    
    command_context : ``CommandContext``
        Respective command context.
    command : ``Category``
        The category to check.
    
    Returns
    -------
    should_show : `bool`
        Whether any command would be shown.
    """
    for command in category.command_instances:
        if await run_command_checks(command_context, command):
            return True
    
    return False


async def list_category_commands(command_context, category):
    """
    Returns the commands which should be shown under the given category.
    
    This function is a coroutine.
    
    command_context : ``CommandContext``
        Respective command context.
    command : ``Category``
        The category to check.
    
    Returns
    -------
    commands : `list` of ``Command``
        The commands to show.
    """
    commands = []
    for command in category.command_instances:
        if await run_command_checks(command_context, command):
            commands.append(command)
    
    return commands


async def run_category_checks(command_context, category):
    """
    Returns whether the given category's checks pass and it is not hidden either.
    
    This function is a coroutine.
    
    command_context : ``CommandContext``
        Respective command context.
    command : ``Category``
        The category to check.
    
    Returns
    -------
    should_show : `bool`
    """
    if category.hidden:
        return False
    
    if category.hidden_if_checks_fail:
        failed_check = await run_checks(category._iter_own_checks(), command_context)
        if (failed_check is not None):
            return False
    
    return True


async def should_show_category(command_context, category):
    """
    Returns whether the given category's should be shown.
    
    This function is a coroutine.
    
    command_context : ``CommandContext``
        Respective command context.
    command : ``Category``
        The category to check.
    
    Returns
    -------
    should_show : `bool`
    """
    if await run_category_checks(command_context, category):
        return await is_category_command_shown(command_context, category)
    
    return False


def iter_sub_commands(command_category, command_name_trace):
    """
    Iterates overt he cub commands of a command or a command category.
    
    This function is a generator.
    
    Parameters
    ----------
    command_category : ``Command`` or ``CommandCategory``
        The command or command category to iterate trough.
    
    Yields
    ------
    command_path_and_function : `str`, ``CommandFunction``
        Path to a command function and itself.
    """
    command_name_trace.append(command_category.display_name)
    command_name_joined = ' '.join(command_name_trace)
    
    command_function = command_category._command_function
    if (command_function is not None):
        yield command_name_joined, command_function
    
    if len(command_name_trace) != MAX_SUB_COMMAND_DEEPNESS:
        command_categories = command_category._command_categories
        if (command_categories is not None):
            for sub_command_category in command_categories:
                yield from iter_sub_commands(sub_command_category, command_name_trace)
    
    del command_name_trace[-1]


def get_sub_commands(command):
    """
    Gets the sub commands and sub command functions of a command.
    
    Parameter
    ---------
    command : ``Command``
        The respective command.
    
    Returns
    -------
    command_paths_and_functions : `list` of `str`, ``CommandFunction``
    Sorted command paths and functions.
    """
    return sorted(iter_sub_commands(command, []))


def get_parameter_name(command_parameter):
    """
    Gets the command parameter's name-
    
    Parameters
    ----------
    command_parameter : ``CommandParameter``
        The parameter.
    
    Returns
    -------
    command_parameter_name : `str`
    """
    parameter_names = []
    for detail in command_parameter._iter_details():
        type_ = detail.type
        if (type_ is not None):
            try:
                parameter_name = PARAMETER_TYPE_TO_NAME[type_]
            except KeyError:
                pass
            else:
                parameter_names.append(parameter_name)
                continue
        
        parameter_name = detail.converter.alternative_type_name
        if parameter_name is None:
            parameter_name = PARAMETER_UNDEFINED
        
        parameter_names.append(parameter_name)
    
    return ' / '.join(parameter_names)


def generate_command_parameter_representation(command_parameter, separator, assigner):
    """
    Generates a command parameter's representation.
    
    Parameters
    ----------
    command_parameter : ``CommandParameter``
        The command parameter to represent.
    separator : `str`
        Separator separating command parameters if needed.
    assigner : `str`
        Assigner to assign command parameters.
    
    Returns
    -------
    representation : `str`
    """
    if command_parameter.is_rest:
        return f'{PARAMETER_OPTIONAL_START}{command_parameter.display_name}{PARAMETER_OPTIONAL_END}'
    
    detail = command_parameter.detail
    if (detail is not None) and (not detail.converter_setting.requires_part):
        return None
    
    if command_parameter.is_kwargs:
        return f'{PARAMETER_OPTIONAL_START}...{assigner}...~1{PARAMETER_OPTIONAL_END}{separator}' \
               f'{PARAMETER_OPTIONAL_START}...{assigner}...~2{PARAMETER_OPTIONAL_END}{separator}' \
               f'...'
    
    parameter_name = get_parameter_name(command_parameter)
    if command_parameter.is_args:
        return f'{PARAMETER_OPTIONAL_START}{parameter_name}~1{PARAMETER_OPTIONAL_END}{separator}' \
               f'{PARAMETER_OPTIONAL_START}{parameter_name}~2{PARAMETER_OPTIONAL_END}{separator}' \
               f'...'
    
    if command_parameter.is_positional:
        return f'{PARAMETER_REQUIRED_START}{parameter_name}{PARAMETER_REQUIRED_END}'
    
    if command_parameter.is_keyword:
        return f'{PARAMETER_OPTIONAL_START}' \
               f'{command_parameter.display_name}{assigner}{parameter_name}' \
               f'{PARAMETER_OPTIONAL_END}'
    
    return None


def generate_command_parameters(command_function):
    """
    Generates command parameter annotations for the given command function.
    
    Parameters
    ----------
    command_function : ``CommandFunction``
        Command function ot generate representation of.
    
    Returns
    -------
    command_parameters : `None` or `str`
        The command function's parameters representation if has any.
    """
    content_parser = command_function._content_parser
    content_parameter_parser = content_parser._content_parameter_parser
    assigner = content_parameter_parser.assigner
    separator = content_parameter_parser.separator
    
    if type(separator) is str:
        if separator in PARAMETER_SEPARATOR_SPACE_AFTER_ONLY:
            parameter_separator = f'{separator} '
        else:
            parameter_separator = f' {separator} '
    else:
        parameter_separator = ' '
    
    if assigner.endswith(' '):
        parameter_assigner = assigner
    else:
        parameter_assigner = f'{assigner} '
    
    parameter_representations = []
    for command_parameter in content_parser._parameters:
        parameter_representation = generate_command_parameter_representation(command_parameter, \
            parameter_separator, parameter_assigner)
        
        if (parameter_representation is not None):
            parameter_representations.append(parameter_representation)
    
    if parameter_representations:
        command_parameters = parameter_separator.join(parameter_representations)
    else:
        command_parameters = None
    
    return command_parameters


def category_sort_key(category):
    """
    Sort key for sorting categories.
    
    Parameters
    ----------
    category : ``Category``
        A command processor's category.
    
    Returns
    -------
    sort_key : `str`
        The categories are sorted based on their display name.
    """
    return category.display_name


class SubterraneanHelpHelp:
    """
    Shows the usage of help of a respective help command.
    
    Attributes
    ----------
    embed_postprocessor : `None` or `callable`
        An embed post-processor, what is called with the autogenerated embeds.
    """
    __slots__ = ('embed_postprocessor', )
    
    def __init__(self, parent):
        """
        Creates a new Subterranean help help instance.
        
        Attributes
        ----------
        parent : ``SubterraneanHelpCommand``
            The parent of the help helper.
        """
        self.embed_postprocessor = parent.embed_postprocessor
    
    async def __call__(self, command_context):
        """
        Returns the respective help command's generated help embed.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The help command's context.
        command : ``Command``
            The command to display it's help.
        
        Returns
        -------
        embed : ``Embed``
        """
        prefix = command_context.prefix
        
        embed = Embed(
            command_context.command.display_name,
            f'Shows the help command of the client.\n'
            f'Try `{prefix}{command_context.command.display_name}` for displaying the categories or the commands\' of them.',
        )
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            embed_postprocessor(command_context, embed)
        
        return embed


class SubterraneanHelpCommand:
    __class_doc__ = (
    """
    A default help command shipped with hata's commands v2 extension.
    
    Usage
    -----
    After you did setup your client, just add the command to your client.
    
    ```py
    from hata import Client
    from hata.ext.commands.helps.subterranean import SubterraneanHelpCommand
    
    Orin = Client('TOKEN',
        extensions = ('commands_v2', 'command_utils'),
        prefix = 'o|',
    )
    
    Orin.commands(SubterraneanHelpCommand(), 'help')
    
    Orin.start()
    ```
    
    You can also add prefix in front of commands, when listing them. Just pass a `prefix` keyword parameter as a string.
    
    If you want to modify each autogenerated embed, then we got you as well! There is an `embed_postprocessor`
    parameter just for this.
    
    The following 2 values are passed to it:
    
    +-----------------------+-------------------------------+
    | Respective name       | Type                          |
    +=======================+===============================+
    | command_context       | ``CommandContext``            |
    +-----------------------+-------------------------------+
    | command_or_category   | ``Command`` or ``Category``   |
    +-----------------------+-------------------------------+
    
    Here is an example using `embed_postprocessor`:
    ```py
    from hata import Client
    from hata.ext.commands.helps.subterranean import SubterraneanHelpCommand
    
    Orin = Client('TOKEN',
        extensions = ('commands_v2', 'command_utils'),
        prefix = 'o|',
    )
    
    def embed_postprocessor(command_context, command_or_category, embed):
        embed.add_thumbnail(command_context.client.avatar_url)
        
        if (embed.color is None):
            if command is help_command:
                color = 0xffffff
            else:
                color = command_context.author.color_at(command_context.guild)
            
            embed.color = color
    
    help_command = Orin.commands(
        SubterraneanHelpCommand(
            embed_postprocessor=embed_postprocessor,
        ),
        'help',
    )
    
    
    # Add a command as well to showcase
    @Orin.commands
    async def ping(client, message):
        return 'pong'
    
    
    Orin.start()
    ```
    
    Note, that `embed_postprocessor` feature is only experimental for now.
    
    Attributes
    ----------
    embed_postprocessor : `None` or `callable`
        An embed post-processor, what is called with the autogenerated embeds.
    prefix : `str`
        Prefix inserted before commands's display name.
    """ ) if DOCS_ENABLED else None
    
    @property
    def __instance_doc__(self):
        """
        Returns the help of the help for the help command.
        
        Returns
        -------
        help_help : ``SubterraneanHelpHelp``
        """
        return SubterraneanHelpHelp(self)
    
    __doc__ = doc_property()
    
    __slots__ = ('embed_postprocessor', 'prefix')
    
    def __new__(cls, color=None, prefix=None, embed_postprocessor=None):
        """
        Creates a new ``SubterraneanHelpCommand`` instance.
        
        Parameters
        ----------
        prefix : `None`, `str`, Optional
            Prefix inserted before commands's display name.
        
        embed_postprocessor : `None` or `callable`, Optional
            An embed post-processor, what is called with the autogenerated embeds.
            
            The following `2` parameters are passed to it:
            
            +-----------------------+-------------------------------+
            | Respective name       | Type                          |
            +=======================+===============================+
            | command_context       | ``CommandContext``            |
            +-----------------------+-------------------------------+
            | command_context       | ``Command`` or ``Category``   |
            +-----------------------+-------------------------------+
        
        Raises
        ------
        TypeError
            If `embed_postprocessor` is given, but not as a `non-async` `callable` accepting `2` parameters.
        """
        if (embed_postprocessor is not None):
            analyzer = CallableAnalyzer(embed_postprocessor)
            min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
            if min_ > 3:
                raise TypeError(f'`embed_postprocessor` should accept `3` parameters: `command_context, '
                    f'command_context`, meanwhile the given one expects at least `{min_!r}`, got '
                    f'`{embed_postprocessor!r}`.')
            
            if (min_ != 2) and (max_ < 2) and (not analyzer.accepts_args()):
                raise TypeError(f'`embed_postprocessor` should accept `2` parameters: `command_context, '
                    f'command_context`, meanwhile the given one expects up to `{max_!r}`, got '
                    f'`{embed_postprocessor!r}`.')
            
            if analyzer.is_async():
                raise TypeError(f'`embed_postprocessor` cannot be async, got `{embed_postprocessor}`.')
        
        if prefix is None:
            prefix = DEFAULT_PREFIX
        else:
            prefix = preconvert_str(prefix, 'prefix', 1, 32)
        
        self = object.__new__(cls)
        self.embed_postprocessor = embed_postprocessor
        self.prefix = prefix
        return self
    
    
    def __repr__(self):
        """Returns the help command's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
        ]
        
        prefix = self.prefix
        if (prefix == DEFAULT_PREFIX):
            field_added = False
        else:
            field_added = True
            repr_parts.append(' prefix=')
            repr_parts.append(repr(prefix))
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            if not field_added:
                repr_parts.append(',')
            
            repr_parts.append(' embed_postprocessor=')
            repr_parts.append(repr(embed_postprocessor))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    async def __call__(self, command_context, name=None):
        """
        Calls the subterranean help command.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The command's context.
        name : `None` or `str`, Optional
            Search term to lookup a category or a command.
        """
        if name is None:
            await self.list_categories(command_context)
            return
        
        if name.isdecimal() and len(name)<21: # do not convert longer names than 64 bit.
            await self.list_category_indexed(command_context, name)
        else:
            await self.lookup_by_name(command_context, name)
    
    
    async def lookup_by_name(self, command_context, name):
        """
        Tries to lookup a command, then a category with the given `name`.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The command's context.
        name : `str`
            Search term to lookup a category or a command.
        """
        name = raw_name_to_display(name)
        
        command_processor = command_context.client.command_processor
        try:
            command = command_processor.command_name_to_command[name]
        except KeyError:
            category = command_processor.category_name_to_category.get(name, None)
            if (category is None) or (not await run_category_checks(command_context, category)):
                await self.command_not_found(command_context, name)
                return
        else:
            if await run_command_checks_in_category(command_context, command):
                await self.show_command(command_context, command)
            else:
                await self.command_not_found(command_context, name)
            return
        
        commands = await list_category_commands(command_context, category)
        if commands:
            command_names = sorted(command.display_name for command in commands)
            await self.list_commands(command_context, command_names, category.display_name)
        else:
            await self.command_not_found(command_context, name)
    
    
    async def show_command(self, command_context, command):
        """
        SHows the given command's description.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            Respective command's context.
        command : ``Command``
            The respective command.
        """
        description = command.description
        if isinstance(description, str):
            await self.show_string_description(command_context, command.display_name, description)
            return
        
        if callable(description):
            embed = await description(command_context)
            embed_type = type(embed)
            if issubclass(embed_type, EmbedBase):
                await Closer(command_context.client, command_context.message.channel, embed,
                    check=partial_func(check_user, command_context.message.author))
                return
            
            if issubclass(embed_type, (tuple, set)) or hasattr(embed_type, '__getitem__') and hasattr(embed_type, '__len__'):
                await Pagination(command_context.client, command_context.message.channel, embed,
                    check=partial_func(check_user, command_context.message.author))
                return
            
            # No more case, go back to default
        
        await self.show_autogenerated_description(command_context, command)
    
    
    async def command_not_found(self, command_context, name):
        """
        Called, when no command or category was found with the given name,
        
        Parameters
        ----------
        command_context : ``CommandContext``
            Command context of the respective
        name : `str`
            The search word.
        """
        embed = Embed('Command not found.',
            f'There is no category or command named as: `{name}`.\n'
            f'Try using `{command_context.prefix}help`',
        )
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            embed_postprocessor(command_context, embed)
        
        await Closer(command_context.client, command_context.message.channel, embed,
            check=partial_func(check_user, command_context.message.author))
    
    
    async def list_categories(self, command_context):
        """
        Lists the categories of the respective client's command preprocessor.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        """
        categories = command_context.client.command_processor.categories
        
        displayable_categories = []
        
        for category in categories:
            if await should_show_category(command_context, category):
                displayable_categories.append(category)
        
        
        
        displayable_categories_length = len(displayable_categories)
        if displayable_categories_length == 1:
            await self.list_category(command_context, displayable_categories[0], display_category_name=False)
            return
        
        category_names = sorted(category.display_name for category in displayable_categories)
        pages = []
        
        if displayable_categories_length == 0:
            pages.append('*[No available category or command]*')
        else:
            page = []
            page_line_count = 0
            for index, category_name in enumerate(category_names, 1):
                page.append(str(index))
                page.append('.: ')
                page.append(category_name)
                
                page_line_count += 1
                
                if page_line_count == MAX_LINE_PER_PAGE:
                    pages.append(''.join(page))
                    page.clear()
                    page_line_count = 0
                else:
                    page.append('\n')
            
            if page_line_count:
                del page[-1]
                pages.append(''.join(page))
        
        page_count = len(pages)
        
        field_name = f'Use `{command_context.prefix}help <category/command>` for more information.'
        
        embeds = [
            Embed('Categories', page).add_field(field_name, f'page {index}/{page_count}')
                for index, page in enumerate(pages, 1)
        ]
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            for embed in embeds:
                embed_postprocessor(command_context, embed)
        
        await Pagination(command_context.client, command_context.message.channel, embeds,
            check=partial_func(check_user, command_context.message.author))
    
    
    async def list_category(self, command_context, category, display_category_name=True):
        """
        Lists the given category's commands.
        
        This command only collects the displayable commands' names, then calls ``.list_commands``.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        category : ``Category``
            The category, which' commands will be listed.
        display_category_name : `bool`, Optional
            Whether the display name of the given category should be displayed. Defaults to `True`.
        """
        commands = await list_category_commands(command_context, category)
        command_names = sorted(command.display_name for command in commands)
        
        if display_category_name:
            category_name = category.display_name
        else:
            category_name = None
        
        await self.list_commands(command_context, command_names, category_name)
    
    
    async def list_commands(self, command_context, command_names, category_name):
        """
        Lists the given commands.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        command_names : `list` of `str`
            The command's names to display.
        category_name : `None or `str`
            The respective category's name.
        """
        pages = []
        page = []
        page_line_count = 0
        prefix = self.prefix
        for command_name in command_names:
            page.append(prefix)
            page.append(' ')
            page.append(command_name)
            
            page_line_count +=1
            
            if page_line_count == MAX_LINE_PER_PAGE:
                pages.append(''.join(page))
                page.clear()
                page_line_count = 0
            else:
                page.append('\n')
        
        if page_line_count:
            del page[-1]
            pages.append(''.join(page))
        
        page_count = len(pages)
        
        field_name = f'Use `{command_context.prefix}help <command>` for more information.'
        
        if category_name is None:
            title = 'Commands'
        else:
            title = f'Commands of {category_name}'
        
        embeds = [
            Embed(title, page).add_field(field_name, f'page {index}/{page_count}')
                for index, page in enumerate(pages, 1)]
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            for embed in embeds:
                embed_postprocessor(command_context, embed)
        
        await Pagination(command_context.client, command_context.message.channel, embeds,
            check=partial_func(check_user, command_context.message.author))
    
    
    async def list_category_indexed(self, command_context, name):
        """
        Tries to find the category with the given `name` (decimal only string) index value.
        
        If the represented index ends up out of bounds, calls ``.lookup_by_name``, if not, then ``.list_commands``.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        name : `str`
            Decimal only string, which might represent a category's index.
        """
        category_index = int(name)
        if category_index == 0:
            await self.lookup_by_name(command_context, name)
            return
        
        categories = iter(sorted(command_context.client.command_processor.categories, key=category_sort_key))
        
        while category_index > 1:
            try:
                category = next(categories)
            except StopIteration:
                await self.lookup_by_name(command_context, name)
                return
            
            if not await should_show_category(command_context, category):
                continue
            
            category_index -= 1
            continue
        
        
        while True:
            try:
                category = next(categories)
            except StopIteration:
                await self.lookup_by_name(command_context, name)
                return
            
            if not await run_category_checks(command_context, category):
                continue
            
            commands = await list_category_commands(command_context, category)
            if not commands:
                continue
            
            break
        
        command_names = sorted(command.display_name for command in commands)
        
        await self.list_commands(command_context, command_names, category.display_name)
    
    
    async def show_string_description(self, command_context, command_name, description):
        """
        Shows the given string description.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        command_name : `str`
            The respective command's display name.
        description : `str`
            The respective command's description.
        """
        if len(description) < 2048:
            embed = Embed(command_name, description)
            
            embed_postprocessor = self.embed_postprocessor
            if (embed_postprocessor is not None):
                embed_postprocessor(command_context, embed)
            
            await Closer(command_context.client, command_context.message.channel, embed,
                check=partial_func(check_user, command_context.message.author))
        
        else:
            description_parts = chunkify(description.splitlines())
            embeds = [Embed(command_name, description_part) for description_part in description_parts]
            
            embed_postprocessor = self.embed_postprocessor
            if (embed_postprocessor is not None):
                for embed in embeds:
                    embed_postprocessor(command_context, embed)
            
            await Pagination(command_context.client, command_context.message.channel, embeds,
                check=partial_func(check_user, command_context.message.author))
    
    
    async def show_autogenerated_description(self, command_context, command):
        """
        Generates, then sends the description of the given description-less command.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The called command's context.
        command : ``Command``
            The respective command.
        """
        prefix = command_context.prefix
        
        description_lines = []
        
        sun_commands = get_sub_commands(command)
        if len(sun_commands) == 0:
            description_lines.append('*no usages*')
        else:
            if len(sun_commands) == 1:
                description_lines.append('Usage:')
            else:
                description_lines.append('Usages:')
            
            for command_path, command_function in sun_commands:
                description_line = ['`', prefix, command_path]
                command_parameters = generate_command_parameters(command_function)
                if command_parameters is not None:
                    description_line.append(' ')
                    description_line.append(command_parameters)
                
                description_line.append('`')
                
                description_lines.append(''.join(description_line))
        
        
        embed = Embed(command.display_name, '\n'.join(description_lines))
        
        embed_postprocessor = self.embed_postprocessor
        if (embed_postprocessor is not None):
            embed_postprocessor(command_context, embed)
        
        await Closer(command_context.client, command_context.message.channel, embed,
            check=partial_func(check_user, command_context.message.author))



__all__ = ()

from re import compile as re_compile, escape as re_escape, M as re_multi_line, S as re_dotall, I as re_ignore_case, \
    match as re_match
from functools import partial as partial_func

from ...backend.utils import FunctionType
from ...backend.analyzer import CallableAnalyzer
from ...backend.export import include

from ...discord.permission.permission import PERMISSION_CAN_SEND_MESSAGES_ALL

from .exceptions import CommandProcessingError, CommandCheckError
from .utils import raw_name_to_display
CheckBase = include('CheckBase')
CommandCheckWrapper = include('CommandCheckWrapper')

SUB_COMMAND_NAME_RP = re_compile('([a-zA-Z0-9_\-]+)\s*')
COMMAND_NAME_RP = re_compile('\s*([^\s]*)\s*', re_multi_line|re_dotall)

async def run_checks(checks, command_context):
    """
    Runs the checks.
    
    This function is coroutine.
    
    Parameters
    ----------
    checks : `generator`
        A generator yielding checks.
    command_context : ``CommandContext``
        The respective command's context.
    
    Returns
    -------
    failed : `None` or ``CheckBase``
        The failed check if any.
    
    Notes
    -----
    Not like ``handle_exception``, this can be called for checks of different commands.
    """
    for check in checks:
        if not await check(command_context):
            return check
    
    return None


async def handle_exception(command_context, exception):
    """
    Handles an exception raised meanwhile processing a command.
    
    This function is a coroutine.
    
    Parameters
    ----------
    command_context : ``CommandContext``
        The respective command context.
    exception : ``BaseException``
        The occurred exception.
    
    Returns
    -------
    invoked : `bool`
        Whether the command was successfully invoked.
        
        If unexpected exception occurs, still returns `True`.
    """
    command_function = command_context.command_function
    if (command_function is not None):
        for error_handler in command_function._iter_error_handlers():
            result = await error_handler(command_context, exception)
            if isinstance(result, int) and result:
                return True
    
    # We can ignore command processing exceptions.
    if isinstance(exception, CommandProcessingError):
        if isinstance(exception, CommandCheckError):
            return command_context.command.hidden_if_checks_fail
    else:
        client = command_context.client
        await client.events.error(client, 'handle_exception', exception)
    
    return True

def get_command_category_trace(command, content, index):
    """
    Gets the sub command trace and command function for the given command.
    
    Parameters
    ----------
    command : ``Command``
        The respective command.
    content : `str`
        A message's content to parse.
    index : `int`
        The starting index from where the content should be parsed from.
    
    Returns
    -------
    command_category_trace : `None` or `tuple` of ``CommandCategory``
        Trace to the actual command.
    command_function : ``CommandFunction`` or `None`
        The command function, which should be called.
    index : `int`
        The index till the command's parameters may start from.
    """
    command_category_name_to_command_category = command.command_category_name_to_command_category
    if (command_category_name_to_command_category is not None):
        trace = []
        end = index
        while True:
            if end == len(content):
                break
            
            parsed = SUB_COMMAND_NAME_RP.match(content, end)
            if (parsed is None):
                break
            
            end = parsed.end()
            part = parsed.group(1)
            name = raw_name_to_display(part)
            
            try:
                command_category = command_category_name_to_command_category[name]
            except KeyError:
                break
            
            trace.append((end, command_category))
            
            sub_commands = command_category._command_categories
            if sub_commands is None:
                break
            
            continue
        
        while trace:
            end, command_category = trace[-1]
            command_function = command_category._command_function
            if (command_function is not None):
                return tuple(trace_element[1] for trace_element in trace), command_function, end
            
            del trace[-1]
            continue
    
    return None, command._command_function, index


def default_precheck(client, message):
    """
    Default check used by the command processor. Filters out every message what's author is a bot account and the
    channels where the client cannot send messages.
    
    Parameters
    ----------
    client : ``Client``
        The client who received the respective message.
    message : ``Message``
        The received message.
    
    Returns
    -------
    should_process : `bool`
    """
    if message.author.is_bot:
        return False
    
    if not message.channel.cached_permissions_for(client)&PERMISSION_CAN_SEND_MESSAGES_ALL:
        return False
    
    return True


def test_precheck(precheck):
    """
    Tests whether the given precheck accepts the expected amount of parameters.
    
    Parameters
    ----------
    precheck : `callable`
        A function, which decides whether a received message should be processed.
        
        The following parameters are passed to it:
        
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | client    | ``Client``    |
        +-----------+---------------+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Should return the following parameters:
        
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    Raises
    ------
    TypeError
        - If `precheck` accepts bad amount of parameters.
        - If `precheck` is async.
    """
    analyzer = CallableAnalyzer(precheck)
    if analyzer.is_async():
        raise TypeError('`precheck` should not be given as `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(f'`precheck` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{precheck!r}`.')
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_args():
                raise TypeError(f'`precheck` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{precheck!r}`.')


def test_error_handler(error_handler):
    """
    Tests whether the given error handler accepts the expected amount of parameters.
    
    Parameters
    ----------
    error_handler : `callable`
        A function, which handles an error and returns whether handled it.
        
        The following parameters are passed to it:
        
        +-------------------+-----------------------+
        | Name              | Type                  |
        +===================+=======================+
        | command_context   | ``CommandContext``    |
        +-------------------+-----------------------+
        | exception         | `BaseException`       |
        +-------------------+-----------------------+
        
        Should return the following parameters:
        
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | handled           | `bool`    |
        +-------------------+-----------+
    
    Raises
    ------
    TypeError
        - If `error_handler` accepts bad amount of parameters.
        - If `error_handler` is not async.
    """
    analyzer = CallableAnalyzer(error_handler)
    if not analyzer.is_async():
        raise TypeError('`error_handler` should be given as `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(f'`error_handler` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{error_handler!r}`.')
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_args():
                raise TypeError(f'`error_handler` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{error_handler!r}`.')


def validate_error_handlers(error_handlers):
    """
    Validates the given error handlers.
    
    Parameters
    ----------
    error_handlers : `None`, `async-callable` or (`list`, `tuple` or `set`) of `async-callable`
        The error_handler(s) to validate.
    
    Returns
    -------
    error_handlers_validated : `None` or `list` of `async-callable`
        The validated error handlers.
    
    Raises
    ------
    TypeError
        - If `error_handler`'s type is incorrect.
        - If `error_handler` accepts bad amount of parameters.
        - If `error_handler` is not async.
    """
    error_handlers_validated = None
    
    if error_handlers is None:
        pass
    elif isinstance(error_handlers, (list, tuple, set)):
        for error_handler in error_handlers:
            test_error_handler(error_handler)
            
            if error_handlers_validated is None:
                error_handlers_validated = []
            
            error_handlers_validated.append(error_handler)
    else:
        test_error_handler(error_handlers)
        
        if error_handlers_validated is None:
            error_handlers_validated = []
        
        error_handlers_validated.append(error_handlers)
    
    return error_handlers_validated


def test_name_rule(rule, rule_name):
    """
    Tests the given name rule and raises if it do not passes any requirements.
    
    Parameters
    ----------
    rule : `None` or `function`
        The rule to test.
        
        A name rule should accept the following parameters:
        
        +-------+-------------------+
        | Name  | Type              |
        +=======+===================+
        | name  | `str`             |
        +-------+-------------------+
        
        Should return the following ones:
        
        +-------+-------------------+
        | Name  | Type              |
        +=======+===================+
        | name  | `str`             |
        +-------+-------------------+
    
    rule_name : `str`
        The name of the given rule.
    
    Raises
    ------
    TypeError
        - If `rule` is not `None` or `function` instance.
        - If `rule` is `async` `function`.
        - If `rule` accepts bad amount of parameters.
        - If `rule` raised exception when `str` was passed to it.
        - If `rule` did not return `str`, when passing `str` to it.
        - If `nullable` is given as `True` and `rule` raised exception when `None` was passed to it.
        - If `nullable` is given as `True` and `rule` did not return `str`, when passing `None` to it.
    """
    if rule is None:
        return
    
    rule_type = rule.__class__
    if (rule_type is not FunctionType):
        raise TypeError(f'`{rule_name}` should have been given as `{FunctionType.__name__}`, got '
            f'{rule_type.__name__}.')
    
    analyzed = CallableAnalyzer(rule)
    if analyzed.is_async():
        raise TypeError(f'`{rule_name}` should have been given as an non async function, got {rule!r}.')
    
    non_reserved_positional_parameter_count = analyzed.get_non_reserved_positional_parameter_count()
    if non_reserved_positional_parameter_count != 1:
        raise TypeError(f'`{rule_name}` should accept `1` non reserved positional parameters, meanwhile it expects '
            f'{non_reserved_positional_parameter_count}.')
    
    if analyzed.accepts_args():
        raise TypeError(f'`{rule_name}` should accept not expect args, meanwhile it does.')
    
    if analyzed.accepts_kwargs():
        raise TypeError(f'`{rule_name}` should accept not expect kwargs, meanwhile it does.')
    
    non_default_keyword_only_parameter_count = analyzed.get_non_default_keyword_only_parameter_count()
    if non_default_keyword_only_parameter_count:
        raise TypeError(f'`{rule_name}` should accept `0` keyword only parameters, meanwhile it expects '
            f'{non_default_keyword_only_parameter_count}.')
    
    try:
        result = rule('test-this-name')
    except BaseException as err:
        raise TypeError(f'Got unexpected exception meanwhile testing the given {rule_name}: {rule!r}.') from err
    
    if (type(result) is not str):
        raise TypeError(f'{rule_name}: {rule!r} did not return `str` instance, meanwhile testing it, got '
            f'{result.__class__.__name__}')


def validate_category_or_command_name(name):
    """
    Validates the given category name.
    
    Parameters
    ----------
    name : `str`
        The name of a category or command.
    
    Returns
    -------
    name : `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` was not given as `str` instance.
    ValueError
        If `name`'s length is out of range [1:128] characters.
    """
    name_type = type(name)
    if name_type is str:
        pass
    elif issubclass(name_type, str):
        name = str(name)
    else:
        raise TypeError(f'Category and command names can be given as `str` instance, got {name_type.__name__}.')
    
    name_length = len(name)
    if (name_length < 1) or (name_length > 128):
        raise ValueError(f'`Category and command name length can be in range [0:128], got {name_length};{name!r}.')
    
    return name


async def prefix_wrapper_async_callable(prefix_factory, re_flags, message):
    """
    Function to execute asynchronous callable prefix.
    
    This function is a coroutine.
    
    Parameters
    ----------
    prefix_factory : `async-callable`
        Async callable returning the prefix.
    re_flags : `int`
        Regex matching flags.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `None` or `str`
        The prefix used by the user. Returned as `None` of parsing failed.
    end : `int`
        The start of the content after the prefix. Returned as `-1` if parsing failed.
    """
    prefix = await prefix_factory(message)
    if isinstance(prefix, str):
        escaped_prefix = re_escape(prefix)
    elif isinstance(prefix, tuple) and (len(prefix) > 0):
        escaped_prefix = '|'.join(re_escape(prefix_part) for prefix_part in prefix)
    else:
        return None, -1
    
    content = message.content
    if content is None:
        prefix = None
        end = -1
    else:
        parsed = re_match(escaped_prefix, content, re_flags)
        if parsed is None:
            prefix = None
            end = -1
        else:
            prefix = parsed.group(0)
            end = parsed.end()
    
    return prefix, end


async def prefix_getter_async_callable(prefix_factory, message):
    """
    Returns a prefix for the message.
    
    This function is a coroutine.
    
    Parameters
    ----------
    prefix_factory : `async-callable`
        Async callable returning the prefix.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `None` or `str`
        The respective prefix for the message. Returns `None` of non could be identified.
    """
    prefix = await prefix_factory(message)
    if isinstance(prefix, str):
        pass
    elif isinstance(prefix, tuple) and (len(prefix) > 0):
        prefix = prefix[0]
    else:
        prefix = None
    
    return prefix


async def prefix_wrapper_sync_callable(prefix_factory, re_flags, message):
    """
    Function to execute not asynchronous callable prefix.
    
    This function is a coroutine.
    
    Parameters
    ----------
    prefix_factory : `async-callable`
        Async callable returning the prefix.
    re_flags : `int`
        Regex matching flags.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `None` or `str`
        The prefix used by the user. Returned as `None` of parsing failed.
    end : `int`
        The start of the content after the prefix. Returned as `-1` if parsing failed.
    """
    prefix = prefix_factory(message)
    if isinstance(prefix, str):
        escaped_prefix = re_escape(prefix)
    elif isinstance(prefix, tuple) and (len(prefix) > 0):
        escaped_prefix = '|'.join(re_escape(prefix_part) for prefix_part in prefix)
    else:
        return None, -1
    
    content = message.content
    if content is None:
        prefix = None
        end = -1
    else:
        parsed = re_match(escaped_prefix, content, re_flags)
        if parsed is None:
            prefix = None
            end = -1
        else:
            prefix = parsed.group(0)
            end = parsed.end()
    
    return prefix, end


async def prefix_getter_sync_callable(prefix_factory, message):
    """
    Returns a prefix for the message.
    
    This function is a coroutine.
    
    Parameters
    ----------
    prefix_factory : `callable`
        Sync callable returning the prefix.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `None` or `str`
        The respective prefix for the message. Returns `None` of non could be identified.
    """
    prefix = prefix_factory(message)
    if isinstance(prefix, str):
        pass
    elif isinstance(prefix, tuple) and (len(prefix) > 0):
        prefix = prefix[0]
    else:
        prefix = None
    
    return prefix


async def prefix_wrapper_regex(re_pattern, message):
    """
    Function to execute asynchronous callable prefix.
    
    This function is a coroutine.
    
    Parameters
    ----------
    re_pattern : `async-callable`
        Async callable returning the prefix.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `None` or `str`
        The prefix used by the user. Returned as `None` of parsing failed.
    end : `int`
        The start of the content after the prefix. Returned as `-1` if parsing failed.
    """
    content = message.content
    if content is None:
        prefix = None
        end = -1
    else:
        parsed = re_pattern.match(content)
        if parsed is None:
            prefix = None
            end = -1
        else:
            prefix = parsed.group(0)
            end = parsed.end()
    
    return prefix, end


async def prefix_getter_static(prefix, message):
    """
    Returns a prefix for the message.
    
    This function is a coroutine.
    
    Parameters
    ----------
    prefix : `str`
        The prefix to return.
    message : ``Message``
        The received message to parse the prefix from.
    
    Returns
    -------
    prefix : `str`
        The respective prefix for the message.
    """
    return prefix


def get_prefix_parser(prefix, prefix_ignore_case):
    """
    Validates whether the given prefix is correct.
    
    prefix :  `str`, `tuple` of `str`, `callable`
        Prefix of the command processor.
        
        Can be given as a normal `callable` or as an `async-callable` as well, which should accept `1` parameter:
        
        +-------------------+---------------+
        | Name              | Type          |
        +===================+===============+
        | message           | ``Message``   |
        +-------------------+---------------+
        
        And return the following value:
        
        +-------------------+---------------------------+
        | Name              | Type                      |
        +===================+===========================+
        | prefix            | `str`, `tuple` of `str`   |
        +-------------------+---------------------------+
    
    Returns
    -------
    prefix_parser : `async-callable`
        Async callable prefix parser.
        
        Accepts the following parameters:
        
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Returns the given values:
        
        +-----------+-------------------+
        | Name      | Type              |
        +===========+===================+
        | prefix    | `None` or `str`   |
        +-----------+-------------------+
        | end       | `int`             |
        +-----------+-------------------+
    
    prefix_getter : `async-callable`
        
        Accepts the following parameters:
        
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Returns the given values:
        
        +-----------+-------------------+
        | Name      | Type              |
        +===========+===================+
        | prefix    | `None` or `str`   |
        +-----------+-------------------+
    
    Raises
    ------
    TypeError
        - Prefix's type is incorrect.
        - Prefix is a callable but accepts bad amount of parameters.
    """
    if prefix_ignore_case:
        re_flags = re_ignore_case
    else:
        re_flags = 0
    
    re_flags |= re_multi_line|re_dotall
    
    if callable(prefix):
        analyzed = CallableAnalyzer(prefix)
        non_reserved_positional_parameter_count = analyzed.get_non_reserved_positional_parameter_count()
        if non_reserved_positional_parameter_count != 1:
            raise TypeError(f'Callable `prefix` should accept `1` non reserved positional parameter, meanwhile it '
                f'accepts: `{non_reserved_positional_parameter_count}`.')
        
        if analyzed.is_async():
            prefix_wrapper_function = prefix_wrapper_async_callable
            prefix_getter_function = prefix_getter_async_callable
        else:
            prefix_wrapper_function = prefix_wrapper_sync_callable
            prefix_getter_function = prefix_getter_sync_callable
        
        prefix_parser = partial_func(prefix_wrapper_function, prefix, re_flags)
        prefix_getter = partial_func(prefix_getter_function, prefix)
    
    else:
        if isinstance(prefix, str):
            escaped_prefix = re_escape(prefix)
        elif isinstance(prefix, tuple):
            if len(prefix) == 0:
                raise ValueError(f'Empty prefix tuple.')
            
            escaped_prefix = '|'.join(re_escape(prefix_part) for prefix_part in prefix)
            prefix = prefix[0]
        else:
            raise TypeError(f'`prefix` can be either given as `callable`, `async-callable`, `str` or as `tuple` of '
                f'`str`, got {prefix.__class__.__name__}.')
        
        compiled_prefix = re_compile(escaped_prefix, re_flags)
        
        prefix_parser = partial_func(prefix_wrapper_regex, compiled_prefix)
        prefix_getter = partial_func(prefix_getter_static, prefix)
    
    return prefix_parser, prefix_getter


def _unwrap_check(check):
    """
    Unwraps the given check if applicable.
    
    Parameters
    ----------
    check : ``CheckBase`` or ``CommandCheckWrapper``
        The check to unwrap.
    
    Returns
    -------
    check : ``CheckBase``
        The unwrapped check.
    
    Raises
    ------
    TypeError
        If `check` is neither ``CheckBase`` nor ``CommandCheckWrapper`` instance.
    """
    if isinstance(check, CheckBase):
        return check
    
    if isinstance(check, CommandCheckWrapper):
        return check._check
    
    raise TypeError(f'`check` can be either `{CheckBase.__name__}` or `{CommandCheckWrapper.__name__}` instance, got '
        f'{check.__class__.__name__}.')


def validate_checks(checks):
    """
    Validates the given checks.
    
    Parameters
    ----------
    checks : `None`, ``CheckBase``, ``CommandCheckWrapper`` or (`list`, `tuple` or `set`) of \
            (``CheckBase`` or ``CommandCheckWrapper``)
        The check(s) to validate.
    
    Returns
    -------
    checks_validated : `None` or `tuple` of ``CheckBase``
        The validated checks.
    
    Raises
    ------
    TypeError
        - If `checks`'s type is incorrect.
        - If a `check`'s type is incorrect.
    """
    checks_validated = None
    
    if checks is None:
        pass
    elif isinstance(checks, (list, tuple, set)):
        for check in checks:
            check = _unwrap_check(check)
            
            if checks_validated is None:
                checks_validated = []
            
            checks_validated.append(check)
        
        checks = tuple(checks_validated)
    else:
        check = _unwrap_check(checks)
        
        checks = (check, )
    
    return checks


def test_unknown_command(unknown_command):
    """
    Tests whether the given unknown command handler accepts the expected amount of parameters.
    
    Parameters
    ----------
    unknown_command : `callable`
        A function, which is called when no command is found.
        
        The following parameters are passed to it:
        
        +---------------+---------------+
        | Name          | Type          |
        +===============+===============+
        | client        | ``Client``    |
        +---------------+---------------+
        | message       | ``Message``   |
        +---------------+---------------+
        | command_name  | `str`         |
        +---------------+---------------+
    
    Raises
    ------
    TypeError
        - If `unknown_command` accepts bad amount of parameters.
        - If `unknown_command` is not async.
    """
    analyzer = CallableAnalyzer(unknown_command)
    if not analyzer.is_async():
        raise TypeError('`unknown_command` should be given as `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 3:
        raise TypeError(f'`unknown_command` should accept `2` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{unknown_command!r}`.')
    
    if min_ != 3:
        if max_ < 3:
            if not analyzer.accepts_args():
                raise TypeError(f'`unknown_command` should accept `2` parameters, meanwhile the given callable expects '
                    f'up to `{max_!r}`, got `{unknown_command!r}`.')



__all__ = ('AttributeUnitBase', 'ClassAttributeUnit', 'FolderedUnit', 'FunctionUnit', 'InstanceAttributeUnit',
    'MAPPED_OBJECTS', 'ModuleUnit', 'ObjectedUnitBase', 'PropertyUnit', 'TypeUnit', 'UnitBase', 'map_module',
    'search_paths')

import sys, re, warnings
from types import FunctionType, BuiltinFunctionType, BuiltinMethodType, MethodType, GetSetDescriptorType, \
    MemberDescriptorType
from difflib import get_close_matches

from ...backend.utils import cached_property, MethodLike, module_property, basemethod, weakmethod
from ...backend.export import include
from ...discord.bases import IconSlot

from .qualpath import QualPath
from .parser import DocString
from .builder_text import serialize_docs_embed_sized, serialize_docs, serialize_docs_source_text, generate_preview_for
from .builder_html import html_serialize_docs

html_serialize_docs_extended = include('html_serialize_docs_extended')

WrapperDescriptorType = object.__eq__.__class__
MethodDescriptorType = int.bit_length.__class__
SlotWrapperType = object.__lt__.__class__

METHOD_TYPES = {
    GetSetDescriptorType,
    BuiltinMethodType,
    MethodType,
    WrapperDescriptorType,
    MethodDescriptorType,
    MethodLike,
    basemethod,
    weakmethod,
}

FUNCTION_TYPES = {
    *METHOD_TYPES,
    FunctionType,
    BuiltinFunctionType,
    BuiltinMethodType,
}

IGNORED_TYPES = {
    type,
    object,
}

IGNORED_CLASS_ATTRIBUTE_TYPES = {
    *METHOD_TYPES,
}

if (SlotWrapperType is not FunctionType) and (SlotWrapperType is not MethodType):
    IGNORED_CLASS_ATTRIBUTE_TYPES.add(SlotWrapperType)

IGNORED_FUNCTIONS = set()

converted = None
for func in (
        object.__delattr__,
        object.__dir__,
        object.__eq__,
        object.__format__,
        object.__ge__,
        object.__getattribute__,
        object.__gt__,
        object.__hash__,
        object.__init__,
        object.__init_subclass__,
        object.__le__,
        object.__lt__,
        object.__ne__,
        object.__new__,
        object.__reduce__,
        object.__reduce_ex__,
        object.__repr__,
        object.__setattr__,
        getattr(object, '__sizeof__', None), # Not present in Pypy
        object.__str__,
        object.__subclasshook__,
        type.__call__,
        type.__delattr__,
        type.__dir__,
        type.__eq__,
        type.__format__,
        type.__ge__,
        type.__getattribute__,
        type.__gt__,
        type.__hash__,
        type.__init__,
        type.__init_subclass__,
        type.__instancecheck__,
        type.__le__,
        type.__lt__,
        type.__ne__,
        type.__new__,
        type.__prepare__,
        type.__reduce__,
        type.__reduce_ex__,
        type.__repr__,
        type.__setattr__,
        getattr(type, '__sizeof__', None), # Not present in Pypy
        type.__str__,
        type.__subclasscheck__,
        type.__subclasses__,
        type.__subclasshook__,
        type.mro,
            ):
    
    if func is None:
        continue
    
    if (func.__class__ in METHOD_TYPES):
        converted = getattr(func, '__func__', None)
        if (converted is not None):
            func = converted
    
    IGNORED_FUNCTIONS.add(func)

del func, converted

PROPERTY_TYPES = {
    property,
    cached_property,
    module_property,
    IconSlot,
}

ATTRIBUTE_TYPES = {
    MemberDescriptorType,
}

IGNORED_CLASS_ATTRIBUTE_NAMES = {
    '__ne__',
    '__module__',
    '__str__',
    '__format__',
    '__lt__',
    '__gt__',
    '__le__',
    '__ge__',
    '__eq__',
    '__reduce__',
    '__reduce_ex__',
    '__setattr__',
    '__getattr__',
    '__delattr__',
    '__slots__',
    '__getattribute__',
    '__repr__',
    '__doc__',
    '__dir__',
    '__sizeof__',
    '__dict__',
    '__new__',
    '__init__',
    '__hash__',
    '__reversed__',
    '__getitem__',
    '__setitem__',
    '__delitem__',
    '__init_subclass__',
    '__subclasshook__',
    '__call__',
    '__class__',
    '__setstate__',
    '__getstate__',
    '__floor__',
    '__mro__',
    '__dictoffset__',
    '__round__',
    '__trunc__',
    '__len__',
    '__bases__',
    '__basicsize__',
    '__name__',
    '__qualname__',
    '__subclasscheck__',
    '__text_signature__',
    '__Weakrefoffset__',
    '__ceil__',
    '__getnewargs__',
    '__contains__',
    '__enter__',
    '__exit__',
    '__slots__',
    '__abstractmethods__',
    '__aenter__',
    '__aexit__',
    '__weakrefoffset__',
    '__instancecheck__',
    '__itemsize__',
    '__subclasses__',
    '__iter__',
    '__weakref__',
}

def map_types_and_functions(obj, references, path, from_type):
    """
    Maps types and functions of the given `obj` to it's `references`.
    
    Parameters
    ----------
    obj : `type` or `module`
        The object to map.
    references : `dict` of (`str`, ``UnitBase``) items
        References of `obj` to it's contained objects.
    path : ``QualPath``
        Tha path of `obj`.
    from_type : `bool`
        Whether `obj` is a `type` instance.
    """
    for attr_name in dir(obj):
        try:
            attr_value = getattr(obj, attr_name)
        except AttributeError:
            continue
        
        attr_value_type = attr_value.__class__
        
        # Process only hashables
        try:
            hash(attr_value)
        except (TypeError, RuntimeError):
            if not from_type:
                continue
            
            if attr_value_type in IGNORED_CLASS_ATTRIBUTE_TYPES:
                continue
            
            if attr_name in IGNORED_CLASS_ATTRIBUTE_NAMES:
                continue
            
            references[attr_name] = ClassAttributeUnit(attr_name, QualPath(path, attr_name))
            continue
        
        # Translate methods
        if attr_value_type in METHOD_TYPES:
            attr_value_func = getattr(attr_value, '__func__', None)
            # Cpython, check. Let built in methods trough.
            if (attr_value_func is not None):
                attr_value = attr_value_func
                attr_value_type = attr_value.__class__
        
        # Check types.
        if issubclass(attr_value_type, type):
            if attr_value in IGNORED_TYPES:
                continue
            
            attr_value_path = QualPath(attr_value.__module__, attr_value.__qualname__)
            
            if attr_value_path.parent == path:
                attr_name = attr_value.__name__
                references[attr_name] = TypeUnit(attr_name, attr_value_path, attr_value)
            
            continue
        
        # Check functions.
        if attr_value_type in FUNCTION_TYPES:
            if attr_value in IGNORED_FUNCTIONS:
                continue
            
            try:
                if from_type:
                    name = attr_value.__qualname__
                else:
                    name = attr_value.__name__
            except AttributeError:
                name = attr_value_type.__name__
            
            if name == '<lambda>':
                name = attr_name
            
            attr_value_module = getattr(attr_value, '__module__', None)
            if attr_value_module is None:
                # Builtin stuff, skip
                continue
            
            attr_value_path = QualPath(attr_value_module, name)
            
            if from_type or attr_value_path.parent == path:
                if from_type:
                    alternative_path = QualPath(path, attr_name)
                else:
                    alternative_path = None
                
                references[attr_name] = FunctionUnit(name, attr_value_path, attr_value, alternative_path)
            
            continue
        
        if not from_type:
            continue
        
        if attr_value_type in PROPERTY_TYPES:
#            for name in ('fget', 'fset', 'fdel'):
#                func = getattr(attr_value,name, None)
#                if (func is not None):
#                    break
#            else:
#                continue
#
#            try:
#                name = func.__name__
#            except AttributeError:
#                name = func.__class__.__name__
#
#            if name == '<lambda>':
#                name = attr_name
            
            references[attr_name] = PropertyUnit(attr_name, QualPath(path, attr_name), attr_value)
            continue
        
        if attr_value_type in ATTRIBUTE_TYPES:
            name = attr_value.__name__
            references[attr_name] = InstanceAttributeUnit(name, QualPath(path, attr_name))
            continue
        
        if attr_value_type in IGNORED_CLASS_ATTRIBUTE_TYPES:
            continue
        
        if attr_name in IGNORED_CLASS_ATTRIBUTE_NAMES:
            continue
        
        references[attr_name] = ClassAttributeUnit(attr_name, QualPath(path, attr_name))
        continue

MAPPED_OBJECTS = {}

class UnitBase:
    """
    Base class for representative units.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the unit.
    path : ``QualPath``
        The path to the unit.
    """
    __slots__ = ('_cache', '_docs', '_docs_parsed', 'name', 'path',)
    
    def __repr__(self):
        """Returns the unit's representation."""
        return f'<{self.__class__.__name__} name={self.name}, path={self.path!s}>'
    
    @property
    def parent(self):
        """
        Returns the parent object of the unit if applicable.
        
        Returns
        -------
        parent : `None` or ``UnitBase`` instance
        """
        parent_path = self.path.parent
        if parent_path:
            parent = MAPPED_OBJECTS.get(parent_path, None)
        else:
            parent = None
        
        return parent
    
    def _get_raw_docstring(self):
        """
        Returns the raw docstring of the represented object.
        
        Returns
        -------
        docstring : `None` or ``DocString``
        """
        return None
    
    @property
    def docs(self):
        """
        Returns the docs of the represented object.
        
        Returns
        -------
        docstring : `None` or ``DocString``
        """
        docs_parsed = self._docs_parsed
        if docs_parsed:
            docstring = self._docs
        else:
            self._docs = docstring = self._get_raw_docstring()
            self._docs_parsed = True
        
        return docstring
    
    @cached_property
    def text(self):
        """
        Renders the docstring of the respective object to one big chunk of string.
        
        Returns
        -------
        docs : `None` or `str`
        """
        docs = self.docs
        if docs is None:
            return None
        
        return serialize_docs(docs)
    
    @cached_property
    def source_text(self):
        """
        Renders the docstring of the respective object to source-like format.
        
        Returns
        -------
        docs : `None` or `str`
        """
        docs = self.docs
        if docs is None:
            return None
        
        return serialize_docs_source_text(docs)
    
    @cached_property
    def embed_sized(self):
        """
        Renders the docstring of the respective object to chunks of string with max size of embed description.
        
        Returns
        -------
        docs : `None` or `list` of `str`
        """
        docs = self.docs
        if docs is None:
            return None
        
        return serialize_docs_embed_sized(docs)
    
    @cached_property
    def html(self):
        """
        Renders the docstring to html.
        
        Returns
        -------
        docs : `None` or `str`
        """
        docs = self.docs
        if docs is None:
            return None
        
        return html_serialize_docs(docs, self)
    
    @cached_property
    def html_extended(self):
        """
        Renders the extended docstring to html.
        
        This html renderer will not only render only the given docstring, but render it with the other related
        docstrings.
        
        Returns
        -------
        html_extended : `None` or `str`
        """
        html_extended, _ = html_serialize_docs_extended(self, True, False)
        return html_extended
    
    @cached_property
    def html_extended_structure(self):
        """
        Creates the extended docstring's html structure.
        
        Returns
        -------
        html_extended_structure : `None` or ``Structure``
        """
        _, html_extended_structure = html_serialize_docs_extended(self, False, True)
        return html_extended_structure
    
    @cached_property
    def html_extended_with_structure(self):
        """
        Renders the extended docstring to html and returns it's structure as well.
        
        Returns
        -------
        html_extended : `None` or `str`
        html_extended_structure : `None` or ``Structure``
        """
        html_extended, html_extended_structure = html_serialize_docs_extended(self, True, True)
        cache = self._cache
        cache['html_extended'] = html_extended
        cache['html_extended_structure'] = html_extended_structure
        return html_extended, html_extended_structure
        
    @cached_property
    def preview(self):
        """
        Generates preview for the docstring.
        
        Returns
        -------
        preview : `None` or `str`
        """
        docs = self.docs
        if docs is None:
            return None
        
        return generate_preview_for(docs)
    
    def lookup_reference(self, reference):
        """
        Returns the closest unit to the given unit.
        
        Returns
        -------
        referred : `None` or ``UnitBase`` instance.
            The referenced unit.
        """
        # Is local reference?
        reference_parts = reference.split('.')
        if not reference_parts:
            return None
        
        reference_parts.reverse()
        
        if not reference_parts[-1]:
            del reference_parts[-1]
            if isinstance(self, FolderedUnit):
                if not reference_parts:
                    return self
                
                object_ = direct_lookup_in(self, reference_parts)
                if (object_ is not None):
                    return object_
            
            parent = MAPPED_OBJECTS.get(self.path.parent, None)
            if parent is None:
                return self
            else:
                return direct_lookup_in(parent, reference_parts)
        
        if isinstance(self, FolderedUnit):
            object_ = lookup_from(self, reference_parts)
            if (object_ is not None):
                return object_
        
        path = self.path.parent
        while path:
            parent = MAPPED_OBJECTS.get(path, None)
            if parent is None:
                break
            
            object_ = lookup_from(parent, reference_parts)
            if (object_ is not None):
                return object_
            
            path = path.parent

def direct_lookup_in(object_, reference_parts):
    """
    Looks up the given reference in a foldered unit directly.
    
    Parameters
    ----------
    object_ : ``UnitBase`` instance
        The folder to lookup up from.
    reference_parts : `list` of `str`
        Reference parts to lookup.
        
        The given `reference_parts` should be reversed from their original state.
    
    Returns
    -------
    object_ : `None` or ``UnitBase``
        The found object if applicable.
    """
    reference_parts = reference_parts.copy()
    
    while reference_parts:
        if not isinstance(object_, FolderedUnit):
            # Cannot move further deep, the reference will not be found, leave.
            object_ = None
            break
        
        reference_part = reference_parts.pop()
        try:
            object_ = object_.references.get(reference_part, None)
        except KeyError:
            # Reference not be found, leave.
            object_ = None
            break
    
    # No more reference part to lookup, return object.
    return object_

def lookup_from(folder, reference_parts):
    """
    Looks up the given reference in the folder and in all of it's sub-folders.
    
    Parameters
    ----------
    folder : ``UnitBase`` instance
        The folder to look up from.
    reference_parts : `list` of `str`
        Reference parts to lookup.
        
        The given `reference_parts` should be reversed from their original state.
    
    Returns
    -------
    object_ : ``UnitBase``
        The unit-base if anything found matching the references.
    """
    if not isinstance(folder, FolderedUnit):
        return None
    
    search_for = reference_parts[-1]
    
    for sub_name, sub_object in folder.references.items():
        if sub_name == search_for:
            if len(reference_parts)  == 1:
                return sub_object
            else:
                object_ = lookup_from(sub_object, reference_parts[:-1])
        else:
            object_ = lookup_from(sub_object, reference_parts)
        
        if object_ is not None:
            return object_
    
    return None

class AttributeUnitBase(UnitBase):
    """
    Base unit class for attributes.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the unit.
    path : ``QualPath``
        The path to the unit.
    """
    __slots__ = ()
    
    def _get_raw_docstring(self):
        """
        Returns the raw docstring of the represented object.
        
        Returns
        -------
        docstring : `None` or ``DocString``
        """
        parent = MAPPED_OBJECTS.get(self.path.parent, None)
        if parent is None:
            return None
        
        docstring = parent.docs
        if docstring is None:
            return None
        
        return docstring.attribute_docstring_for(self.name)


class InstanceAttributeUnit(AttributeUnitBase):
    """
    Represents an instance attribute.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the attribute.
    path : ``QualPath``
        The path to the attribute.
    """
    __slots__ = ()
    def __new__(cls, name, path):
        """
        Creates a new instance attribute unit from the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the attribute
        path : ``QualPath``
            The path to the attribute.
        """
        self = object.__new__(cls)
        self.name = name
        self.path = path
        self._cache = {}
        self._docs = None
        self._docs_parsed = False
        MAPPED_OBJECTS[path] = self
        
        return self


class ClassAttributeUnit(AttributeUnitBase):
    """
    Represents a class attribute
    
    Attributes
    ----------
    name : `str`
        The name of the attribute.
    path : ``QualPath``
        The path to the attribute.
    """
    __slots__ = ()
    def __new__(cls, name, path):
        """
        Creates a new class attribute unit from the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the attribute
        path : ``QualPath``
            The path to the attribute.
        """
        self = object.__new__(cls)
        self.name = name
        self.path = path
        self._cache = {}
        self._docs = None
        self._docs_parsed = False
        
        MAPPED_OBJECTS[path] = self
        
        return self


class ObjectedUnitBase(UnitBase):
    """
    Base class for docstrings, which store their represented object.

    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the unit.
    path : ``QualPath``
        The path to the unit.
    object : `property-like`
        The represented property.
    """
    __slots__ = ('object', )
    
    def _get_raw_docstring(self):
        """
        Returns the raw docstring of the represented object.
        
        Returns
        -------
        docstring : `None` or ``DocString``
        """
        docstring = getattr(self.object, '__doc__', 'None')
        if docstring is None:
            return None
        
        if type(docstring) is not str:
            return None
        
        return DocString(docstring, self.path)


class PropertyUnit(ObjectedUnitBase):
    """
    Represents a property.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the unit.
    path : ``QualPath``
        The path to the unit.
    object : `property-like`
        The represented property.
    """
    __slots__ = ()
    def __new__(cls, name, path, obj):
        """
        Creates a new property unit from the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the property.
        path : ``QualPath``
            The path to the property.
        obj : `property-like`
            The represented property.
        """
        self = object.__new__(cls)
        self.name = name
        self.path = path
        self.object = obj
        self._cache = {}
        self._docs = None
        self._docs_parsed = False
        
        MAPPED_OBJECTS[path] = self
        
        return self


class FunctionUnit(ObjectedUnitBase):
    """
    Represents a function or method.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the function.
    path : ``QualPath``
        The path to the function.
    object : `function-like`
        The represented function.
    """
    __slots__= ()
    def __new__(cls, name, path, obj, alternative_path):
        """
        Creates a new property unit from the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the function.
        path : ``QualPath``
            The path to the function.
        obj : `property-like`
            The represented function.
        alternative_path : `None` or ``QualPath``
            Alternative path to the function.
            
            Methods might show up at multiple places, so more access path can be added to them.
        """
        try:
            self = MAPPED_OBJECTS[path]
        except KeyError:
            self = object.__new__(cls)
            self.name = name
            self.path = path
            self.object = obj
            self._cache = {}
            self._docs = None
            self._docs_parsed = False
            
            MAPPED_OBJECTS[path] = self
        
        if (alternative_path is not None) and (alternative_path != path):
            MAPPED_OBJECTS[alternative_path] = self
        
        return self


class FolderedUnit(ObjectedUnitBase):
    """
    Represents a foldered unit.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the unit.
    path : ``QualPath``
        The path to the unit.
    references : `dict` of (`str`, `UnitBase`) items
        The references of the object to the objects contained by itself. AKA it's folder.
    object : `property-like`
        The represented unit.
    """
    __slots__ = ('references',)
    
    def __repr__(self):
        """Returns the foldered unit's representation."""
        return (f'<{self.__class__.__name__} name={self.name!r}, path={self.path!s}, reference count='
            f'{len(self.references)!r}>')


class TypeUnit(FolderedUnit):
    """
    Represents a type.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the type.
    path : ``QualPath``
        The path to the type.
    references : `dict` of (`str`, `UnitBase`) items
        The references of the type to the objects contained by itself. AKA it's folder.
    object : `type`
        The represented type.
    """
    __slots__ = ()
    def __new__(cls, name, path, obj):
        """
        Creates a new type unit from the given `obj`.
        
        Parameters
        ----------
        name : `str`
            The name of the type.
        path : ``QualPath``
            The path to the type.
        obj : `type`
            The represented type.
        """
        references = {}
        map_types_and_functions(obj, references, path, True)
        
        self = object.__new__(cls)
        self.name = name
        self.path = path
        self.references = references
        self.object = obj
        self._cache = {}
        self._docs = None
        self._docs_parsed = False
        
        MAPPED_OBJECTS[path] = self
        
        return self


class ModuleUnit(FolderedUnit):
    """
    Represents a module.
    
    Attributes
    ----------
    _cache : `dict`
        Cache used by cached properties.
    _docs : `None` or ``DocString``
        The processed docstring of the represented unit.
    _docs_parsed : `bool`
        Whether the represented unit's docstring was already processed.
    name : `str`
        The name of the module.
    path : ``QualPath``
        The path to the module.
    references : `dict` of (`str`, `UnitBase`) items
        The references of the module to the objects contained by itself. AKA it's folder.
    object : `module`
        The represented module.
    """
    __slots__= ()
    def __new__(cls, name, path, obj, modules):
        """
        Creates a new module unit from the given `obj`.
        
        Parameters
        ----------
        name : `str`
            The name of the module.
        path : ``QualPath``
            The path to the module.
        obj : `module`
            The represented module.
        """
        references = {}
        
        base_path_pattern = re.compile(f'{re.escape(str(path))}\.([^\.]+)')
        for module_object in modules:
            parsed = base_path_pattern.fullmatch(module_object.__name__)
            if parsed is None:
                continue
            
            sub_module_name = parsed.group(1)
            references[sub_module_name] = \
                ModuleUnit(sub_module_name, path / sub_module_name, module_object, modules)
        
        map_types_and_functions(obj, references, path, False)
        
        self = object.__new__(cls)
        self.name = name
        self.path = path
        self.references = references
        self.object = obj
        self._cache = {}
        self._docs = None
        self._docs_parsed = False
        
        MAPPED_OBJECTS[path] = self
        
        return self

def map_module(module_name):
    """
    Maps the given module by it's name and returns the root.
    
    Attributes
    ----------
    module_name : `str`
        The module's name.
    
    Returns
    -------
    root : ``ModulePath``
    
    Raises
    ------
    LookupError
        No loaded module is named as `module_name`.
    """
    try:
        return MAPPED_OBJECTS[module_name]
    except KeyError:
        pass
    
    try:
        main_module_object = sys.modules[module_name]
    except KeyError:
        raise LookupError(f'No loaded module is named as {module_name!r}') from None
    
    # Invalidate search cache
    CachedSearcher._cache_valid = False
    
    module_name_pattern = re.compile(f'{re.escape(module_name)}(?:\..*)?')
    modules = []
    
    for name, module_object in sys.modules.items():
        if (module_name_pattern.fullmatch(name) is not None):
            modules.append(module_object)
    
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        root = ModuleUnit(module_name, QualPath(module_name), main_module_object, modules)
    
    return root


def qual_path_sort_key(qual_path):
    """
    Sort key to sort ``QualPath`` instances.
    
    Parameters
    ----------
    qual_path : ``QualPath``
        The respective path.
    
    Returns
    -------
    sort_key : `str`
    """
    return str(qual_path)


class CachedSearcher:
    """
    Cached unit path searcher.
    
    Class Attributes
    ----------------
    _cached_relations : `dict` of (``QualPath``, (``QualPath`` or `list` of ``QualPath``)) items
        Path shortening, path relations used when translating found patches back.
    _cached_possibilities : `list` of `str`
        The cached possibilities.
    _cache_valid : `bool`
        Whether the searcher cache is valid.
    """
    _cache_valid = True
    _cached_relations = {}
    _cached_possibilities = []
    
    @classmethod
    def _prepare(cls):
        """
        Prepares the cache of the searcher.
        
        Returns
        -------
        possibilities : `list` of `str`
            The cached possibilities.
        relations : `dict` of (``QualPath``, (``QualPath`` or `list` of ``QualPath``)) items
            Path shortening, path relations used when translating found patches back.
        """
        cached_relations = cls._cached_relations
        cached_possibilities = cls._cached_possibilities
        if not cls._cache_valid:
            cached_relations.clear()
            cached_possibilities.clear()
            for path in MAPPED_OBJECTS:
                path_parts = path.parts
                for starter in range(len(path_parts)):
                    local_path = '.'.join(path_parts[starter:])
                    
                    try:
                        collected_paths = cached_relations[local_path]
                    except KeyError:
                        cached_relations[local_path] = path
                        cached_possibilities.append(local_path)
                    else:
                        if isinstance(collected_paths, QualPath):
                            cached_relations[local_path] = [collected_paths, path]
                        else:
                            collected_paths.append(path)
            
            for value in cached_relations.values():
                if isinstance(value, QualPath):
                    continue
                
                value.sort(key=qual_path_sort_key)
        
        return cached_possibilities, cached_relations
    
    def __call__(self, value, limit=100):
        """
        Returns all the found units with the given value in their path. Case insensitive.
        
        Parameters
        ----------
        value : `str`
            The query string.
        
        Returns
        -------
        paths : `list` of ``QualPath``
        """
        possibilities, relations = self._prepare()
        
        matcheds = get_close_matches(value, possibilities, n=limit, cutoff=0.60)
        
        paths = []
        collected = set()
        
        for matched in matcheds:
            path = relations[matched]
            if isinstance(path, QualPath):
                if path in collected:
                    continue
                
                collected.add(path)
                paths.append(path)
            else:
                for path in path:
                    if path in collected:
                        continue
                    
                    collected.add(path)
                    paths.append(path)
        
        del paths[limit:]
        
        return paths


search_paths = CachedSearcher()



__all__ = ()

from os.path import split as split_path
from collections import deque

from ...backend.utils import to_json
from ...backend.export import include
from ...backend.formdata import Formdata

from ..core import MESSAGES, CHANNELS, GUILDS, USERS, STICKERS, SCHEDULED_EVENTS
from ..message import Message, MessageReference, MessageRepr
from ..user import ClientUserBase
from ..channel import ChannelText, ChannelStage
from ..embed import EmbedBase
from ..utils import random_id
from ..bases import maybe_snowflake_pair, maybe_snowflake, maybe_snowflake_token_pair
from ..guild import Guild, GuildDiscovery
from ..oauth2 import Achievement
from ..role import Role
from ..stage import Stage
from ..webhook import Webhook
from ..emoji import Emoji, parse_reaction
from ..sticker import Sticker
from ..scheduled_event import ScheduledEvent


ComponentBase = include('ComponentBase')
ComponentType = include('ComponentType')
ComponentRow = include('ComponentRow')

def get_components_data(components, is_edit):
    """
    Gets component data from the given components.
    
    Parameters
    ----------
    components : `None`, ``ComponentBase``, (`set`, `list`) of \
            (``ComponentBase``, (`set`, `list`) of ``ComponentBase``)
        Components to be attached to a message.
    is_edit : `bool`
        Whether the processed `components` fields are for message edition. At this case passing `None` will
        remove them.
    
    Returns
    -------
    component_datas : `None` ot `list` of (`dict` of (`str`, `Any`) items)
        The generated data if any.
    
    Raises
    ------
    TypeError
        - If `components` was not given neither as `None`, ``ComponentBase``, (`list`, `tuple`) of ``ComponentBase``
            instances.
    AssertionError
        - If `components` contains a non ``ComponentBase`` element.
    """
    
    # Components check order:
    # 1.: None -> None || []
    # 2.: Ellipsis -> None || Ellipsis
    # 2.: ComponentBase -> [component.to_data()]
    # 3.: (list, tuple) of ComponentBase, (list, tuple) of ComponentBase -> [component.to_data(), ...] / None
    # 4.: raise
    
    if components is None:
        if is_edit:
            component_datas = []
        else:
            component_datas = None
    
    elif components is ...:
        if is_edit:
            component_datas = ...
        else:
            component_datas = None
    
    else:
        if isinstance(components, ComponentBase):
            if components.type is not ComponentType.row:
                components = ComponentRow(components)
                
            component_datas = [components.to_data()]
        elif isinstance(components, (list, tuple)):
            component_datas = None
            
            for component in components:
                if isinstance(component, ComponentBase):
                    if component.type is not ComponentType.row:
                        component = ComponentRow(component)
                
                elif isinstance(component, (list, tuple)):
                    component = ComponentRow(*component)
                
                else:
                    raise TypeError(f'`components` contains a non `{ComponentBase.__name__}` or as `list`, `tuple` of '
                        f'`{ComponentBase.__name__}` instances, got {components.__class__.__name__}')
                
                if component_datas is None:
                    component_datas = []
                
                component_datas.append(component.to_data())
                continue
        
        else:
            raise TypeError(f'`components` can be given as `{ComponentBase.__name__}` or as `list`, `tuple` of '
                f'(`{ComponentBase.__name__}` or (`list`, `tuple`) of `{ComponentBase.__name__}`), '
                f'got {components.__class__.__name__}')
    
    return component_datas


def validate_message_to_delete(message):
    """
    Validates a message to delete.
    
    Parameters
    ----------
    message : ``Message``, ``MessageReference``, ``MessageRepr``, `tuple` (`int`, `int`)
        The message to validate for deletion.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier where the message is.
    message_id : `int`
        The message's identifier.
    message : `None` or ``Message``
        The referenced message if found.
    
    Raises
    ------
    TypeError
        If message was not given neither as ``Message``, ``MessageReference``, ``MessageRepr``, neither as
        `tuple` (`int`, `int`).
    """
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    else:
        if isinstance(message, MessageRepr):
            channel_id = message.channel_id
            message_id = message.id
        elif isinstance(message, MessageReference):
            channel_id = message.channel_id
            message_id = message.message_id
        else:
            snowflake_pair = maybe_snowflake_pair(message)
            if snowflake_pair is None:
                raise TypeError(f'`message` should have be given as `{Message.__name__}` or as '
                    f'`{MessageRepr.__name__}`, `{MessageReference.__name__}`, or as `tuple` of (`int`, `int`), '
                    f'got {message.__class__.__name__}.')
            
            channel_id, message_id = snowflake_pair
        
        message = MESSAGES.get(message, None)
    
    return channel_id, message_id, message


def create_file_form(data, file):
    """
    Creates a `multipart/form-data` form from the message's data and from the file data. If there is no files to
    send, will return `None` to tell the caller, that nothing is added to the overall data.
    
    Parameters
    ----------
    data : `dict` of `Any`
        The data created by the ``.message_create`` method.
    file : `dict` of (`file-name`, `io`) items, `list` of (`file-name`, `io`) elements, tuple (`file-name`, `io`), `io`
        The files to send.
    
    Returns
    -------
    form : `None` or `Formdata`
        Returns a `Formdata` of the files and from the message's data. If there are no files to send, returns `None`
        instead.
    
    Raises
    ------
    ValueError
        If more than `10` file is registered to be sent.
    
    Notes
    -----
    Accepted `io` types with check order are:
    - ``BodyPartReader`` instance
    - `bytes`, `bytearray`, `memoryview` instance
    - `str` instance
    - `BytesIO` instance
    - `StringIO` instance
    - `TextIOBase` instance
    - `BufferedReader`, `BufferedRandom` instance
    - `IOBase` instance
    - ``AsyncIO`` instance
    - `async-iterable`
    
    Raises `TypeError` at the case of invalid `io` type.
    
    There are two predefined data types specialized to send files:
    - ``ReuBytesIO``
    - ``ReuAsyncIO``
    
    If a buffer is sent, then when the request is done, it is closed. So if the request fails, we would not be
    able to resend the file, except if we have a data type, what instead of closing on `.close()` just seeks to
    `0` (or later if needed) on close, instead of really closing instantly. These data types implement a
    `.real_close()` method, but they do `real_close` on `__exit__` as well.
    """
    form = Formdata()
    form.add_field('payload_json', to_json(data))
    files = []
    
    # checking structure
    
    # case 1 dict like
    if hasattr(type(file), 'items'):
        files.extend(file.items())
    
    # case 2 tuple => file, filename pair
    elif isinstance(file, tuple):
        files.append(file)
    
    # case 3 list like
    elif isinstance(file, (list, deque)):
        for element in file:
            if type(element) is tuple:
                name, io = element
            else:
                io = element
                name = ''
            
            if not name:
                #guessing name
                name = getattr(io, 'name', '')
                if name:
                    _, name = split_path(name)
                else:
                    name = str(random_id())
            
            files.append((name, io),)
    
    #case 4 file itself
    else:
        name = getattr(file, 'name', '')
        #guessing name
        if name:
            _, name = split_path(name)
        else:
            name = str(random_id())
        
        files.append((name, file),)
    
    # checking the amount of files
    # case 1 one file
    if len(files) == 1:
        name, io = files[0]
        form.add_field('file', io, filename=name, content_type='application/octet-stream')
    # case 2, no files -> return None, we should use the already existing data
    elif len(files) == 0:
        return None
    # case 3 maximum 10 files
    elif len(files) < 11:
        for index, (name, io) in enumerate(files):
            form.add_field(f'file{index}s', io, filename=name, content_type='application/octet-stream')
    
    # case 4 more than 10 files
    else:
        raise ValueError('You can send maximum 10 files at once.')
    
    return form


def add_file_to_message_data(message_data, file, contains_content):
    """
    Adds files to the message data creating a form data if applicable.
    
    Parameters
    ----------
    message_data : `dict` of (`str`, `Any`) items
        The message's payload to send.
    file : `None` or `dict` of (`file-name`, `io`) items, `list` of (`file-name`, `io`) elements, \
            tuple (`file-name`, `io`), `io`
        The files to send.
    contains_content : `bool`
        Whether the message already contains any content.
    
    Returns
    -------
    message_data : `None` or `dict`, `Formdata`
        Returns a `Formdata` if the message contains attachments, `dict` if contains any content and `None` if
        not.
    
    Raises
    ------
    ValueError
        If more than `10` file is registered to be sent.
    """
    if (file is None):
        if not contains_content:
            message_data = None
    else:
        form = create_file_form(message_data, file)
        if (form is None) and (not contains_content):
            message_data = None
        else:
            message_data = form
    
    return message_data


def validate_content_and_embed(content, embed, is_edit):
    """
    Validates the given content and embed fields of a message creation or edition.
    
    Parameters
    ----------
    content : `str`, ``EmbedBase`` or `Any`, Optional
        The content of the message.
        
        
        If given as ``EmbedBase`` instance, then the message's embeds will be edited with it.
    embed : `None`, ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
        The new embedded content of the message. By passing it as `None`, you can remove the old.
        
        > If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `AssertionError` is
        raised.
    is_edit : `bool`
        Whether the processed `content` and `embed` fields are for message edition. At this case passing `None` will
        remove them.
    
    Returns
    -------
    content : `Ellipsis`, `None`, `str`
        The message's content.
    embed : `Ellipsis`, `None`, ``EmbedBase``, (`list` or `tuple`) of ``EmbedBase``
        The messages embeds.
    
    Raises
    ------
    TypeError
        If `embed` was not given neither as ``EmbedBase`` nor as `list` or `tuple` of ``EmbedBase`` instances.
    AssertionError
        - If `embed` contains a non ``EmbedBase`` element.
        - If both `content` and `embed` fields are embeds.
    """
    # Embed check order:
    # 1.: None
    # 2.: Ellipsis -> None || Ellipsis
    # 3.: Embed : -> embed || [embed]
    # 4.: list of Embed -> embed[0] || embed[:10] or None
    # 5.: raise
    if embed is None:
        pass
    
    elif embed is ...:
        if not is_edit:
            embed = None
    
    elif isinstance(embed, EmbedBase):
        embed = [embed]
    
    elif isinstance(embed, (list, tuple)):
        if embed:
            if __debug__:
                for embed_element in embed:
                    if not isinstance(embed_element, EmbedBase):
                        raise AssertionError(f'`embed` was given as a `list` or `tuple`, but it\'s it contains a non '
                            f'`{EmbedBase.__name__}` instance element, got {embed_element.__class__.__name__}.')
            
            embed = embed[:10]
        else:
            embed = None
    
    else:
        raise TypeError(f'`embed` was not given as `{EmbedBase.__name__}` instance, neither as a `list`  or `tuple` of '
            f'{EmbedBase.__name__} instances, got {embed.__class__.__name__}.')
    
    # Content check order:
    # 1.: None -> None || ''
    # 2.: Ellipsis -> None || Ellipsis
    # 3.: str
    # 4.: Embed -> embed = content || [content]
    # 5.: list of Embed -> embed = content[0] || content[:10]
    # 6.: object -> str(content)
    
    if content is None:
        if is_edit:
            content = ''
    
    elif content is ...:
        if not is_edit:
            content = None
    
    elif isinstance(content, str):
        pass
    
    elif isinstance(content, EmbedBase):
        if __debug__:
            if (embed is not (... if is_edit else None)):
                raise AssertionError(f'Multiple embeds were given, got content={content!r}, embed={embed!r}.')
        
        embed = [content]
        
        if is_edit:
            content = ...
        else:
            content = None
    
    else:
        # Check for list of embeds as well.
        if isinstance(content, (list, tuple)):
            if content:
                for element in content:
                    if isinstance(element, EmbedBase):
                        continue
                    
                    is_list_of_embeds = False
                    break
                else:
                    is_list_of_embeds = True
            else:
                is_list_of_embeds = False
        else:
            is_list_of_embeds = False
        
        if is_list_of_embeds:
            if __debug__:
                if (embed is not (... if is_edit else None)):
                    raise AssertionError(f'Multiple embeds were given, got content={content!r}, embed={embed!r}.')
            
            embed = content[:10]
            
            if is_edit:
                content = ...
            else:
                content = None
        else:
            content = str(content)
    
    return content, embed


def get_channel_id(channel, channel_type):
    """
    Gets the channel's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : `channel_type`, `int`
        The channel, or it's identifier.
    channel_type : `type`
        The expected channel type of `channel`.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, channel_type):
        channel_id = channel.id
    
    else:
        channel_id = maybe_snowflake(channel)
        if channel_id is None:
            raise TypeError(f'The channel can be either given as `{channel_type.__name__}` or as `int` instance, '
                f'got {channel.__class__.__name__}.')
    
    return channel_id


def get_guild_and_guild_text_channel_id(channel):
    """
    Gets guild text channel identifier and it's guild from the given text channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``ChannelText``, `int`
        The channel, or it's identifier.
    
    Returns
    -------
    guild : ``Guild`` or `None`
        The respective guild if any found.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    while True:
        if isinstance(channel, ChannelText):
            channel_id = channel.id
            guild = channel.guild
            break
        
        channel_id = maybe_snowflake(channel)
        if (channel_id is not None):
            try:
                channel = CHANNELS[channel_id]
            except KeyError:
                guild = None
                break
            
            if isinstance(channel, ChannelText):
                guild = channel.guild
                break
        
        raise TypeError(f'`channel` can be either given as `{ChannelText.__name__}` or as `int` instance, '
            f'got {channel.__class__.__name__}.')
    
    return guild, channel_id


def get_guild_id_and_channel_id(channel, channel_type):
    """
    Gets the channel's and it's guild's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : `channel_type` or `tuple` (`int`, `int`)
        The role, or `guild-id`, `role-id` pair.
    channel_type : `type`
        The expected type of `channel`.
    
    Returns
    -------
    snowflake_pair : `None` or `tuple` (`int`, `int`)
        The channel's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, channel_type):
        guild = channel.guild
        if guild is None:
            snowflake_pair = None
        else:
            snowflake_pair = guild.id, channel.id
    else:
        snowflake_pair = maybe_snowflake_pair(channel)
        if snowflake_pair is None:
            raise TypeError(f'`channel` can be given as `{channel_type.__name__}`, or as '
                f'`tuple` (`int`, `int`), got {channel.__class__.__name__}.')
        
    return snowflake_pair


def get_channel_and_id(channel, channel_type):
    """
    Gets thread channel and it's identifier from the given thread channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``ChannelThread``, `int`
        The channel, or it's identifier.
    channel_type : `type`
        The expected type of `channel`.
    
    Returns
    -------
    channel : ``channel_type``, `None`
        The channel if found.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    while True:
        if isinstance(channel, channel_type):
            channel_id = channel.id
            break
        
        channel_id = maybe_snowflake(channel)
        if (channel_id is not None):
            try:
                channel = CHANNELS[channel_id]
            except KeyError:
                channel = None
                break
            
            if isinstance(channel, channel_type):
                break
        
        raise TypeError(f'`channel` can be either given as `{channel_type.__name__}` or as `int` instance, '
            f'got {channel.__class__.__name__}.')
    
    return channel, channel_id


def get_stage_channel_id(stage):
    """
    Gets stage's channel's identifier from the given stage or of it's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``ChannelStage``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    if isinstance(stage, Stage):
        channel_id = stage.channel.id
    elif isinstance(stage, ChannelStage):
        channel_id = stage.id
    else:
        channel_id = maybe_snowflake(stage)
        if channel_id is None:
            raise TypeError(f'`stage` can be given as `{Stage.__name__}`, `{ChannelStage.__name__}`, or as '
                f'int` instance, got {stage.__class__.__name__}.')
    
    return channel_id


def get_user_id(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if isinstance(user, ClientUserBase):
        user_id = user.id
    
    else:
        user_id = maybe_snowflake(user)
        if user_id is None:
            raise TypeError(f'`user` can be either given as `{ClientUserBase.__name__}` or as `int` instance, '
                f'got {user.__class__.__name__}.')
    
    return user_id


def get_user_and_id(user):
    """
    Gets user and it's identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user : ``ClientUserBase`` or `None`
        The user if found.
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    while True:
        if isinstance(user, ClientUserBase):
            user_id = user.id
            break
        
        user_id = maybe_snowflake(user)
        if (user_id is not None):
            try:
                user = USERS[user_id]
            except KeyError:
                user = None
                break
            
            if isinstance(user, ClientUserBase):
                break
        
        raise TypeError(f'`user` can be either given as `{ClientUserBase.__name__}` or as `int` instance, '
            f'got {user.__class__.__name__}.')
        
    return user, user_id


def get_user_id_nullable(user):
    """
    Gets user identifier from the given user or of it's identifier. The user can be given as `None`. At that case
    `user_id` will default to `0`.
    
    Parameters
    ----------
    user : `None`, ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if user is None:
        user_id = 0
    
    elif isinstance(user, ClientUserBase):
        user_id = user.id
    
    else:
        user_id = maybe_snowflake(user)
        if user_id is None:
            raise TypeError(f'`user` can be either given as `{ClientUserBase.__name__}` or as `int` instance, '
                f'got {user.__class__.__name__}.')
    
    return user_id


def get_guild_id(guild):
    """
    Gets the guild's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(f'`guild` can be given as `{Guild.__name__}` or `int` instance, got '
                f'{guild.__class__.__name__}.')
    
    return guild_id


def get_guild_and_id(guild):
    """
    Gets the guild and it's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild : ``Guild`` or `None`
        The guild if found.
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(f'`guild` can be given as `{Guild.__name__}` or `int` instance, got '
                f'{guild.__class__.__name__}.')
        
        guild = GUILDS.get(guild_id, None)
    
    return guild, guild_id


def get_guild_discovery_and_id(guild):
    """
    Gets the guild discovery and it's identifier from the given guild or it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild : ``GuildDiscovery`` or `None`
        The guild or the
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
        guild_discovery = None
    elif isinstance(guild, GuildDiscovery):
        guild_id = guild.guild.id
        guild_discovery = guild
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(f'`guild` can be given as `{Guild.__name__}`, `{GuildDiscovery.__name__}` or `int` '
                f'instance, got {guild.__class__.__name__}.')
        
        guild_discovery = None
    
    return guild_discovery, guild_id


def get_achievement_id(achievement):
    """
    Gets the achievement identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(f'`achievement` can be given as `{Achievement.__name__}` or `int` instance, got '
                f'{achievement.__class__.__name__}.')
    
    return achievement_id


def get_achievement_and_id(achievement):
    """
    Gets the achievement and it's identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement : ``Achievement``, `None`
        The achievement if found.
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(f'`achievement` can be given as `{Achievement.__name__}` or `int` instance, got '
                f'{achievement.__class__.__name__}.')
        
        achievement = None
    
    return achievement, achievement_id


def get_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 2.: MessageRepr
    # 3.: MessageReference
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif isinstance(message, MessageRepr):
        channel_id = message.channel_id
        message_id = message.id
    elif isinstance(message, MessageReference):
        channel_id = message.channel_id
        message_id = message.message_id
    elif message is None:
        raise TypeError('`message` was given as `None`. Make sure to use `Client.message_create` with giving '
            'content and using a cached channel.')
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(f'`message` can be given as `{Message.__name__}` or as '
                f'`{MessageRepr.__name__}`, `{MessageReference.__name__}` or as `tuple` of (`int`, `int`), got '
                f'`{message.__class__.__name__}`.')
        
        channel_id, message_id = snowflake_pair
    
    return channel_id, message_id


def get_role_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `int`
        The role, or it's identifier.
    
    Returns
    -------
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
    else:
        role_id = maybe_snowflake(role)
        if role_id is None:
            raise TypeError(f'`role` can be given as `{Role.__name__}` or `int` instance, got '
                f'{role.__class__.__name__}.')
    
    return role_id


def get_guild_id_and_role_id(role):
    """
    Gets the role's and it's guild's identifier from the given role or of a `guild-id`, `role-id` pair.
    
    Parameters
    ----------
    role : ``Role`` or `tuple` (`int`, `int`)
        The role, or `guild-id`, `role-id` pair.
    
    Returns
    -------
    snowflake_pair : `None` or `tuple` (`int`, `int`)
        The role's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        guild = role.guild
        if guild is None:
            snowflake_pair = None
        else:
            snowflake_pair = guild.id, role.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(role)
        if snowflake_pair is None:
            raise TypeError(f'`role` can be given as `{Role.__name__}`, or as `tuple` (`int`, `int`), got '
                f'{role.__class__.__name__}.')
    
    return snowflake_pair


def get_webhook_id(webhook):
    """
    Gets the webhook's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        webhook_id = webhook.id
    else:
        webhook_id = maybe_snowflake(webhook)
        if webhook_id is None:
            raise TypeError(f'`webhook` can be given as `{Webhook.__name__}` or `int` instance, got '
                f'{webhook.__class__.__name__}.')
    
    return webhook_id


def get_webhook_and_id(webhook):
    """
    Gets the webhook and it's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook : ``Webhook`` or `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            break
        
        webhook_id = maybe_snowflake(webhook)
        if (webhook_id is not None):
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
            
            raise TypeError(f'`webhook` can be given as `{Webhook.__name__}` or `int` instance, got '
                f'{webhook.__class__.__name__}.')
    
    return webhook, webhook_id


def get_webhook_id_token(webhook):
    """
    Gets the webhook's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        snowflake_token_pair = webhook.id, webhook.token
    else:
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is None):
            raise TypeError(f'`webhook` can be given either as `{Webhook.__name__}` or as `tuple` (`int`, `str`), '
                f'got {webhook.__class__.__name__}.')
    
    return snowflake_token_pair


def get_webhook_and_id_token(webhook):
    """
    Gets the webhook, it's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook : ``Webhook`` or `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            webhook_token = webhook.token
            break
        
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is not None):
            webhook_id, webhook_token = snowflake_token_pair
            
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
        
        raise TypeError(f'`webhook` can be given either as `{Webhook.__name__}` or as `tuple` (`int`, `str`), '
            f'got {webhook.__class__.__name__}.')
    
    return webhook, webhook_id, webhook_token


def get_reaction(emoji):
    """
    Gets the reaction form of the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji`` or `str`
        The emoji to get it's reaction form of.
    
    Returns
    -------
    as_reaction : `str`
        The emoji's reaction form.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        as_reaction = emoji.as_reaction
    elif isinstance(emoji, str):
        as_reaction = emoji
    else:
        raise TypeError(f'`emoji` can be given as ``{Emoji.__class__}`` or as `str` instance, got '
            f'{emoji.__class__.__name__}.')
    
    return as_reaction


def get_emoji_from_reaction(emoji):
    """
    Gets emoji from the given emoji or reaction string.
    
    Parameters
    ----------
    emoji : ``Emoji`` or `str`
        The emoji or reaction to get the emoji from.
    
    Returns
    -------
    emoji : ``Emoji``
        The emoji itself.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    ValueError
        The given `emoji` is not a valid reaction.
    """
    if isinstance(emoji, Emoji):
        pass
    elif isinstance(emoji, str):
        emoji = parse_reaction(emoji)
        if emoji is None:
            raise ValueError(f'The given `emoji` is not a valid reaction, got {emoji!r}.')
    else:
        raise TypeError(f'`emoji` can be given as ``{Emoji.__class__}`` or as `str` instance, got '
            f'{emoji.__class__.__name__}.')
    
    return emoji


def get_guild_id_and_emoji_id(emoji):
    """
    Gets the emoji's and it's guild's identifier from the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji`` or `tuple` (`int`, `int`)
        The emoji, or `guild-id`, `emoji-id` pair.
    
    Returns
    -------
    snowflake_pair : `None` or `tuple` (`int`, `int`)
        The emoji's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        guild = emoji.guild
        if guild is None:
            snowflake_pair = None
        else:
            snowflake_pair = guild.id, emoji.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(emoji)
        if snowflake_pair is None:
            raise TypeError(f'`emoji` can be given as `{Emoji.__name__}`, or as `tuple` (`int`, `int`), got '
                f'{emoji.__class__.__name__}.')
    
    return snowflake_pair


def get_sticker_and_id(sticker):
    """
    Gets sticker and it's identifier from the given sticker or of it's identifier.
    
    Parameters
    ----------
    sticker : ``Sticker``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker : ``Sticker`` or `None`
        The sticker if found.
    sticker_id : `int`
        The sticker's identifier.
    
    Raises
    ------
    TypeError
        If `sticker`'s type is incorrect.
    """
    while True:
        if isinstance(sticker, Sticker):
            sticker_id = sticker.id
            break
        
        sticker_id = maybe_snowflake(sticker)
        if (sticker_id is not None):
            try:
                sticker = STICKERS[sticker_id]
            except KeyError:
                sticker = None
                break
            
            if isinstance(sticker, Sticker):
                break
        
        raise TypeError(f'`sticker` can be either given as `{Sticker.__name__}` or as `int` instance, '
            f'got {sticker.__class__.__name__}.')
        
    return sticker, sticker_id


def get_scheduled_event_id(scheduled_event):
    """
    Gets the scheduled event's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `int`
        The scheduled event, or it's identifier.
    
    Returns
    -------
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        scheduled_event_id = scheduled_event.id
    else:
        scheduled_event_id = maybe_snowflake(scheduled_event)
        if scheduled_event_id is None:
            raise TypeError(f'`scheduled_event` can be given as `{ScheduledEvent.__name__}` or `int` instance, got '
                f'{scheduled_event.__class__.__name__}.')
    
    return scheduled_event_id


def get_scheduled_event_and_id(scheduled_event):
    """
    Gets the scheduled event's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `int`
        The scheduled event, or it's identifier.
    
    Returns
    -------
    scheduled_event : `None` or ``ScheduledEvent``
        The scheduled event.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        scheduled_event =  scheduled_event
        scheduled_event_id = scheduled_event.id
    else:
        scheduled_event_id = maybe_snowflake(scheduled_event)
        if scheduled_event_id is None:
            raise TypeError(f'`scheduled_event` can be given as `{ScheduledEvent.__name__}` or `int` instance, got '
                f'{scheduled_event.__class__.__name__}.')
        
        scheduled_event = SCHEDULED_EVENTS.get(scheduled_event_id, None)
    
    return  scheduled_event, scheduled_event_id
